# -*- coding: utf-8 -*-
################################################################################
#
#  pyCatapult -- Python Interface to ECRS Catapult
#  Copyright © 2016-2017 Lance Edgar
#
#  This file is part of pyCatapult.
#
#  pyCatapult is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  pyCatapult is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  pyCatapult.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Schema

The data models contained here represent the SMS database tables which have
been mapped thus far.  Certainly there are more tables which have not yet been
mapped.

For convenience, here is a list of the models:

* :class:`Customer` (table ``Customers``)
* :class:`InventoryItem` (table ``v_InventoryMaster``)
* :class:`Prefix` (table ``PrefixDescriptions``)
* :class:`Suffix` (table ``SuffixDescriptions``)
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy


class ModelBase(object):
    """
    Base class for all data models.
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    def __repr__(self):
        # TODO: Seems like this could be more elegant...  It apparently is
        # necessary due to the mismatch between column and attribute names.
        mapper = orm.object_mapper(self)
        keys = []
        for keycol in mapper.primary_key:
            found = False
            for attr in mapper.column_attrs.values():
                if len(attr.columns) == 1 and attr.columns[0] is keycol:
                    keys.append('{}={}'.format(attr.key, repr(getattr(self, attr.key))))
                    found = True
                    break
            assert found
        assert keys
        return "{}({})".format(self.__class__.__name__, ', '.join(keys))


Base = declarative_base(cls=ModelBase)


def getset_factory(collection_class, proxy):
    """
    Get/set factory for SQLAlchemy association proxy attributes.
    """
    def getter(obj):
        if obj is None:
            return None
        return getattr(obj, proxy.value_attr)

    def setter(obj, val):
        setattr(obj, proxy.value_attr, val)

    return getter, setter


@six.python_2_unicode_compatible
class Store(Base):
    """
    Represents a store.
    """
    __tablename__ = 'Stores'

    pk = sa.Column('STO_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('STO_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    number = sa.Column('STO_Number', sa.String(length=4), nullable=False)
    name = sa.Column('STO_Name', sa.String(length=60), nullable=False)

    def __str__(self):
        return self.name or ''


class Tax(Base):
    """
    Represents a tax rate
    """
    __tablename__ = 'TaxMain'

    pk = sa.Column('TAX_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('TAX_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    name = sa.Column('TAX_Name', sa.String(length=30), nullable=False)
    receipt_alias = sa.Column('TAX_ReceiptAlias', sa.String(length=30), nullable=False)
    rate_percent = sa.Column('TAX_RatePercent', sa.Numeric(precision=7, scale=4), nullable=True)


@six.python_2_unicode_compatible
class Department(Base):
    """
    Represents a department record.
    """
    __tablename__ = 'Departments'

    pk = sa.Column('DPT_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the department record.
    """)

    number = sa.Column('DPT_Number', sa.Integer(), nullable=False, doc="""
    Presumably (?) unique number for the department.
    """)

    name = sa.Column('DPT_Name', sa.String(length=30), nullable=False, doc="""
    Name of the department.
    """)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Vendor(Base):
    """
    Represents a vendor (aka supplier?) record.
    """
    __tablename__ = 'Vendor'

    pk = sa.Column('VEN_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('VEN_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    code = sa.Column('VEN_Code', sa.String(length=10), nullable=False, doc="""
    Presumably (?) unique code for the vendor.
    """)

    company_name = sa.Column('VEN_CompanyName', sa.String(length=25), nullable=False, doc="""
    Name for the vendor's company.
    """)

    def __str__(self):
        return (self.company_name or '').strip()


@six.python_2_unicode_compatible
class Brand(Base):
    """
    Represents a brand record.
    """
    __tablename__ = 'Brands'

    pk = sa.Column('BRD_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the brand record.
    """)

    name = sa.Column('BRD_Name', sa.String(length=30), nullable=False, doc="""
    Name of the brand.
    """)

    def __str__(self):
        return (self.name or '').strip()


# TODO: add OrderUnitProfiles?


class WeightProfile(Base):
    """
    Represents a weight profile (``WeightProfiles``) record.
    """
    __tablename__ = 'WeightProfiles'

    pk = sa.Column('WGT_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    name = sa.Column('WGT_Name', sa.String(length=30), nullable=False)

    tare_prompt = sa.Column('WGT_TarePrompt', sa.Boolean(), nullable=False)


@six.python_2_unicode_compatible
class DiscountProfile(Base):
    """
    Represents a discount profile record.
    """
    __tablename__ = 'DiscountProfiles'

    pk = sa.Column('DIS_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('DIS_Description', sa.String(length=30), nullable=False)

    max_quantity = sa.Column('DIS_MaxQty', sa.Integer(), nullable=False)

    item_limit = sa.Column('DIS_ItemLimit', sa.Integer(), nullable=True)

    def __str__(self):
        return (self.description or '').strip()

    @property
    def formula(self):
        if self.formulas:
            if len(self.formulas) > 1:
                raise ValueError("discount has multiple ({}) formulas".format(len(self.formulas)))
            return self.formulas[0]


class DiscountFormulaAttribute(Base):
    """
    Represents a discount formula, more or less..
    """
    __tablename__ = 'DiscountFormulaAttributes'
    __table_args__ = (
        sa.ForeignKeyConstraint(['FOX_DIS_FK'], ['DiscountProfiles.DIS_PK']),
    )

    pk = sa.Column('FOX_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    discount_fk = sa.Column('FOX_DIS_FK', sa.Integer(), nullable=False)
    discount = orm.relationship(DiscountProfile, backref=orm.backref('formulas'))

    min_quantity = sa.Column('FOX_MinQty', sa.Numeric(precision=12, scale=3), nullable=False)

    type_override = sa.Column('FOX_TypeOverride', sa.Integer(), nullable=False)

    amount_override = sa.Column('FOX_AmountOverride', sa.Numeric(precision=16, scale=4), nullable=True)


@six.python_2_unicode_compatible
class ComboGroupProfile(Base):
    """
    Represents a "combo group"
    """
    __tablename__ = 'ComboGroupProfiles'

    pk = sa.Column('CGP_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('CGP_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    name = sa.Column('CGP_Name', sa.String(length=30), nullable=False)

    items = association_proxy('_items', 'item', getset_factory=getset_factory)

    def __str__(self):
        return (self.name or '').strip()


@six.python_2_unicode_compatible
class ComboGroupItem(Base):
    """
    Defines the item(s) which comprise a combo group..?
    """
    __tablename__ = 'ComboGroupItems'
    __table_args__ = (
        sa.ForeignKeyConstraint(['CGI_CGP_FK', 'CGI_CGP_CFK'], ['ComboGroupProfiles.CGP_PK', 'ComboGroupProfiles.CGP_CPK']),
        sa.ForeignKeyConstraint(['CGI_INV_FK', 'CGI_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
    )

    pk = sa.Column('CGI_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('CGI_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    group_fk = sa.Column('CGI_CGP_FK', sa.Integer(), nullable=True)
    group_cfk = sa.Column('CGI_CGP_CFK', sa.Integer(), nullable=True)
    group = orm.relationship(ComboGroupProfile, backref=orm.backref('_items'))

    item_fk = sa.Column('CGI_INV_FK', sa.Integer(), nullable=True)
    item_cfk = sa.Column('CGI_INV_CFK', sa.Integer(), nullable=True)
    item = orm.relationship('StockInventory')

    def __str__(self):
        return (self.name or '').strip()


@six.python_2_unicode_compatible
class MarketingCategory(Base):
    """
    Represents a marketing category (attribute?)
    """
    __tablename__ = 'MarketingCategories'

    pk = sa.Column('MCA_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('MCA_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('MCA_Description', sa.String(length=30), nullable=True)

    def __str__(self):
        return (self.description or '').strip()


@six.python_2_unicode_compatible
class WebCustomSource(Base):
    """
    Represents a web custom source (?)
    """
    __tablename__ = 'WebCustomSource'

    pk = sa.Column('WCS_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('WCS_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('WCS_Description', sa.String(length=30), nullable=False)

    def __str__(self):
        return self.description.strip()


class StockInventoryBasePrice(Base):
    """
    Represents a base price (``v_StockInventoryBasePrice``) record.
    """
    __tablename__ = 'v_StockInventoryBasePrice'
    __table_args__ = (
        sa.ForeignKeyConstraint(['SIB_DIS_FK'], ['DiscountProfiles.DIS_PK']),
    )

    pk = sa.Column('SIB_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    base_price = sa.Column('SIB_BasePrice', sa.Numeric(precision=16, scale=4), nullable=False)

    prompt_for_price = sa.Column('SIB_PromptForPrice', sa.Boolean(), nullable=False)

    preset_price = sa.Column('SIB_PresetPrice', sa.Boolean(), nullable=False)

    ideal_margin = sa.Column('SIB_IdealMargin', sa.Numeric(precision=16, scale=4), nullable=True)

    discount_fk = sa.Column('SIB_DIS_FK', sa.Integer(), nullable=True)
    discount = orm.relationship(DiscountProfile)

    discount_description = sa.Column('DIS_Description', sa.String(length=30), nullable=True)


class PowerFieldDataInventory1(Base):
    """
    Represents a record in the inventory power field 1 lookup table.
    """
    __tablename__ = 'PowerFieldDataInventory1'

    pk = sa.Column('PI1_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('PI1_Description', sa.String(length=30), nullable=False)


class Image(Base):
    """
    Represents an image stored in the Catapult DB.
    """
    __tablename__ = 'Images'

    pk = sa.Column('IMG_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('IMG_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    source = sa.Column('IMG_Source', sa.LargeBinary(), nullable=False)
    type = sa.Column('IMG_Type', sa.String(length=3), nullable=True)
    meta = sa.Column('IMG_Meta', sa.Text(), nullable=True)
    category = sa.Column('IMG_Category', sa.Integer(), nullable=True)
    hash = sa.Column('IMG_Hash', sa.String(length=40), nullable=True)


@six.python_2_unicode_compatible
class StockInventory(Base):
    """
    Represents a "stock inventory" item...?
    """
    __tablename__ = 'StockInventory'
    __table_args__ = (
        sa.ForeignKeyConstraint(['INV_DPT_FK'], ['Departments.DPT_PK']),
        sa.ForeignKeyConstraint(['INV_BRD_FK'], ['Brands.BRD_PK']),
        sa.ForeignKeyConstraint(['INV_CGP_FK', 'INV_CGP_CFK'], ['ComboGroupProfiles.CGP_PK', 'ComboGroupProfiles.CGP_CPK']),
    )

    pk = sa.Column('INV_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('INV_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    scancode = sa.Column('INV_ScanCode', sa.String(length=14), nullable=False)

    department_fk = sa.Column('INV_DPT_FK', sa.Integer(), nullable=True)
    department = orm.relationship(Department, backref=orm.backref('stock_inventory_items'))

    brand_fk = sa.Column('INV_BRD_FK', sa.Integer(), nullable=True)
    brand = orm.relationship(Brand)

    name = sa.Column('INV_Name', sa.String(length=30), nullable=True)

    size = sa.Column('INV_Size', sa.String(length=20), nullable=True)

    memo = sa.Column('INV_Memo', sa.String(length=254), nullable=True)

    combo_group_fk = sa.Column('INV_CGP_FK', sa.Integer(), nullable=True)
    combo_group_cfk = sa.Column('INV_CGP_CFK', sa.Integer(), nullable=True)
    combo_group = orm.relationship(ComboGroupProfile)

    country_of_origin_label = sa.Column('INV_Cool', sa.String(length=100), nullable=True)

    discontinued = sa.Column('INV_Discontinued', sa.Boolean(), nullable=True)

    health_attributes = association_proxy('_health_attributes', 'attribute',
                                          getset_factory=getset_factory)

    taxes = association_proxy('_taxes', 'tax', getset_factory=getset_factory)

    websources = association_proxy('_websources', 'websource', getset_factory=getset_factory)

    images = association_proxy('_images', 'image', getset_factory=getset_factory)

    primary_ordering_info = orm.relationship(
        'OrderingInfo',
        primaryjoin='and_(OrderingInfo.item_fk == StockInventory.pk, OrderingInfo.item_cfk == StockInventory.cpk, OrderingInfo.primary == 1)',
        uselist=False)

    def __str__(self):
        return (self.name or '').strip()

    def has_health_attribute(self, name):
        """
        Convenience method, returns boolean to indicate if the given name is
        found among the item's Health Attributes.
        """
        return name in [attr.description for attr in self.health_attributes]


class StockInventoryLinkImage(Base):
    """
    Represents a link between item and image.
    """
    __tablename__ = 'StockInventoryLinkImages'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ILI_INV_FK', 'ILI_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
        sa.ForeignKeyConstraint(['ILI_IMG_FK', 'ILI_IMG_CFK'], ['Images.IMG_PK', 'Images.IMG_CPK']),
    )

    item_fk = sa.Column('ILI_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item_cfk = sa.Column('ILI_INV_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item = orm.relationship(StockInventory, backref=orm.backref('_images'))

    image_fk = sa.Column('ILI_IMG_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    image_cfk = sa.Column('ILI_IMG_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    image = orm.relationship(Image)


class StockInventoryWebCustomSource(Base):
    """
    Represents a link between item and web custom source (?)
    """
    __tablename__ = 'WebCustomSourceLinkStockInventory'
    __table_args__ = (
        sa.ForeignKeyConstraint(['WCI_INV_FK', 'WCI_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
        sa.ForeignKeyConstraint(['WCI_WCS_FK', 'WCI_WCS_CFK'], ['WebCustomSource.WCS_PK', 'WebCustomSource.WCS_CPK']),
    )

    inv_fk = sa.Column('WCI_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    inv_cfk = sa.Column('WCI_INV_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item = orm.relationship(StockInventory, backref=orm.backref('_websources'))

    wcs_fk = sa.Column('WCI_WCS_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    wcs_cfk = sa.Column('WCI_WCS_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    websource = orm.relationship(WebCustomSource)


class StockInventoryLocal(Base):
    """
    Represents local things for a stock inventory item.
    """
    __tablename__ = 'StockInventoryLocal'
    __table_args__ = (
        sa.ForeignKeyConstraint(['SIL_INV_FK', 'SIL_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
        sa.ForeignKeyConstraint(['SIL_STO_FK', 'SIL_STO_CFK'], ['Stores.STO_PK', 'Stores.STO_CPK']),
    )

    item_fk = sa.Column('SIL_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item_cfk = sa.Column('SIL_INV_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    # TODO: pretty sure uselist=False for backref wouldn't work for multi-store
    # setup...or rather, it would at least issue warnings
    item = orm.relationship(StockInventory, backref=orm.backref('local', uselist=False))

    store_fk = sa.Column('SIL_STO_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    store_cfk = sa.Column('SIL_STO_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    store = orm.relationship(Store)

    location = sa.Column('SIL_Location', sa.String(length=30), nullable=True)
    last_received = sa.Column('SIL_LastReceived', sa.DateTime(), nullable=True)
    last_sold = sa.Column('SIL_LastSold', sa.DateTime(), nullable=True)

    discontinued = sa.Column('SIL_Discontinued', sa.Boolean(), nullable=True)
    date_discontinued = sa.Column('SIL_DateDiscontinued', sa.DateTime(), nullable=True)


@six.python_2_unicode_compatible
class HealthAttribute(Base):
    """
    Represents a health attribute.
    """
    __tablename__ = 'HealthAttributes'

    pk = sa.Column('HAT_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('HAT_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('HAT_Description', sa.String(length=30), nullable=True)

    def __str__(self):
        return self.description or ''


class StockInventoryLinkHealthAttribute(Base):
    """
    Links an item with a health attribute.
    """
    __tablename__ = 'StockInventoryLinkHealthAttributes'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ILH_INV_FK'], ['StockInventory.INV_PK']),
        sa.ForeignKeyConstraint(['ILH_HAT_FK', 'ILH_HAT_CFK'], ['HealthAttributes.HAT_PK', 'HealthAttributes.HAT_CPK']),
    )

    inv_fk = sa.Column('ILH_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    hat_fk = sa.Column('ILH_HAT_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    hat_cfk = sa.Column('ILH_HAT_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    item = orm.relationship(StockInventory, backref=orm.backref('_health_attributes'))
    attribute = orm.relationship(HealthAttribute)


@six.python_2_unicode_compatible
class InventoryItem(Base):
    """
    Represents an inventory item (``v_InventoryMaster``) record.
    """
    __tablename__ = 'v_InventoryMaster'
    __table_args__ = (
        sa.ForeignKeyConstraint(['inv_pk', 'inv_cpk'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
        sa.ForeignKeyConstraint(['inv_dpt_fk'], ['Departments.DPT_PK']),
        sa.ForeignKeyConstraint(['inv_brd_fk'], ['Brands.BRD_PK']),
        sa.ForeignKeyConstraint(['inv_wgt_fk'], ['WeightProfiles.WGT_PK']),
        sa.ForeignKeyConstraint(['sib_dis_fk'], ['DiscountProfiles.DIS_PK']),
    )

    pk = sa.Column('inv_pk', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('inv_cpk', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item = orm.relationship(
        StockInventory,
        backref=orm.backref(
            '_master',
            uselist=False))

    scancode = sa.Column('inv_scancode', sa.String(length=14), nullable=False, doc="""
    Scan code for the item.
    """)

    name = sa.Column('inv_name', sa.String(length=30), nullable=True, doc="""
    Name (description) for the item.
    """)

    size = sa.Column('inv_size', sa.String(length=20), nullable=True, doc="""
    Size string for the item.
    """)

    suggested_retail = sa.Column('inv_suggestedretail', sa.Numeric(precision=16, scale=4), nullable=False)

    memo = sa.Column('inv_memo', sa.String(length=254), nullable=True, doc="""
    Memo string for the item.
    """)

    department_fk = sa.Column('inv_dpt_fk', sa.Integer(), nullable=False)
    department = orm.relationship(Department, backref=orm.backref('_inventory_master_items'))

    department_name = sa.Column('dpt_name', sa.String(length=30), nullable=False, doc="""
    Department name for the product.
    """)

    department_number = sa.Column('dpt_number', sa.Integer(), nullable=False, doc="""
    Department number for the product.
    """)

    brand_fk = sa.Column('inv_brd_fk', sa.Integer(), nullable=True)
    brand = orm.relationship(Brand)

    brand_name = sa.Column('brd_name', sa.String(length=30), nullable=True, doc="""
    Brand name for the product.
    """)

    base_price = sa.Column('sib_baseprice', sa.Numeric(precision=16, scale=4), nullable=False, doc="""
    Base price for the item.
    """)

    ideal_margin = sa.Column('sib_idealmargin', sa.Numeric(precision=16, scale=4), nullable=True)

    prompt_for_price = sa.Column('sib_promptforprice', sa.Boolean(), nullable=False)

    weight_fk = sa.Column('inv_wgt_fk', sa.Integer(), nullable=True)
    weight = orm.relationship(WeightProfile)

    weight_name = sa.Column('wgt_name', sa.String(length=30), nullable=True)

    discount_fk = sa.Column('sib_dis_fk', sa.Integer(), nullable=True)
    discount = orm.relationship(DiscountProfile)

    location = sa.Column('inv_location', sa.String(length=30), nullable=True)

    discontinued = sa.Column('inv_discontinued', sa.Boolean(), nullable=True)

    # TODO: rename these?
    pi1_description = sa.Column('pi1_description', sa.String(length=30), nullable=True)
    pi2_description = sa.Column('pi2_description', sa.String(length=30), nullable=True)
    pi3_description = sa.Column('pi3_description', sa.String(length=30), nullable=True)
    pi4_description = sa.Column('pi4_description', sa.String(length=30), nullable=True)

    # TODO: rename these?
    inv_powerfield1 = sa.Column('inv_powerfield1', sa.String(length=30), nullable=True)
    inv_powerfield2 = sa.Column('inv_powerfield2', sa.String(length=30), nullable=True)
    inv_powerfield3 = sa.Column('inv_powerfield3', sa.String(length=30), nullable=True)
    inv_powerfield4 = sa.Column('inv_powerfield4', sa.String(length=30), nullable=True)

    quantity_in_order_unit = sa.Column('ord_quantityinorderunit', sa.Float(), nullable=True)

    vendor_code = sa.Column('ven_code', sa.String(length=10), nullable=True)
    supplier_stock_number = sa.Column('ord_supplierstocknumber', sa.String(length=15), nullable=True)

    alternate_scancode = sa.Column('asc_scancode', sa.String(length=14), nullable=True)
    alternate_receipt_alias = sa.Column('asc_receiptalias', sa.String(length=32), nullable=True)
    alternate_quantity = sa.Column('asc_quantity', sa.Numeric(precision=12, scale=3), nullable=True)

    orderinfo = orm.relationship(
        'OrderingInfo',
        primaryjoin='and_(OrderingInfo.item_fk == InventoryItem.pk, OrderingInfo.item_cfk == InventoryItem.cpk, OrderingInfo.primary == 1)',
        foreign_keys='OrderingInfo.item_fk, OrderingInfo.item_cfk',
        uselist=False)

    country_of_origin_label = sa.Column('inv_cool', sa.String(length=100), nullable=True)

    def __str__(self):
        return (self.name or '').strip()


class ItemAlternate(Base):
    """
    Represents an "Alternate ID" (``AdditionalScanCodes``) record for the item.
    """
    __tablename__ = 'AdditionalScanCodes'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ASC_INV_FK', 'ASC_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
    )

    pk = sa.Column('ASC_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    item_fk = sa.Column('ASC_INV_FK', sa.Integer(), nullable=False)
    item_cfk = sa.Column('ASC_INV_CFK', sa.Integer(), nullable=False)
    item = orm.relationship(
        StockInventory,
        backref=orm.backref('alternates'))

    scancode = sa.Column('ASC_ScanCode', sa.String(length=14), nullable=False)

    quantity = sa.Column('ASC_Quantity', sa.Numeric(precision=12, scale=3), nullable=True)

    receipt_alias = sa.Column('ASC_ReceiptAlias', sa.String(length=32), nullable=True)

    default = sa.Column('ASC_Default', sa.Boolean(), nullable=False)


class OrderingInfo(Base):
    """
    Represents an "ordering info" (``OrderingInfo``) record.
    """
    __tablename__ = 'OrderingInfo'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ORD_INV_FK', 'ORD_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
        sa.ForeignKeyConstraint(['ORD_VEN_FK', 'ORD_VEN_CFK'], ['Vendor.VEN_PK', 'Vendor.VEN_CPK']),
    )

    pk = sa.Column('ORD_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('ORD_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    item_fk = sa.Column('ORD_INV_FK', sa.Integer(), nullable=False)
    item_cfk = sa.Column('ORD_INV_CFK', sa.Integer(), nullable=False)
    item = orm.relationship(
        StockInventory,
        backref=orm.backref('ordering_infos'))

    vendor_fk = sa.Column('ORD_VEN_FK', sa.Integer(), nullable=False)
    vendor_cfk = sa.Column('ORD_VEN_CFK', sa.Integer(), nullable=False)
    vendor = orm.relationship(Vendor)

    supplier_stock_number = sa.Column('ORD_SupplierStockNumber', sa.String(length=15), nullable=False)

    primary = sa.Column('ORD_Primary', sa.Boolean(), nullable=False)

    quantity_in_order_unit = sa.Column('ORD_QuantityInOrderUnit', sa.Numeric(precision=12, scale=3), nullable=False)
    minimum_order = sa.Column('ORD_MinimumOrder', sa.Numeric(precision=12, scale=3), nullable=True)
    suggested_multiple = sa.Column('ORD_SuggestedMultiple', sa.Numeric(precision=12, scale=3), nullable=True)
    allow_override_of_multiple = sa.Column('ORD_AllowOverrideOfMultiple', sa.Boolean(), nullable=True)

    discontinued = sa.Column('ORD_Discontinued', sa.Boolean(), nullable=True)
    date_discontinued = sa.Column('ORD_DateDiscontinued', sa.DateTime(), nullable=True)


class StockInventoryLinkTax(Base):
    """
    Represents a link between a stock inventory item and a tax rate.
    """
    __tablename__ = 'StockInventoryLinkTax'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ILT_INV_FK'], ['StockInventory.INV_PK']),
        sa.ForeignKeyConstraint(['ILT_TAX_FK', 'ILT_TAX_CFK'], ['TaxMain.TAX_PK', 'TaxMain.TAX_CPK']),
    )

    sto_fk = sa.Column('ILT_STO_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    inv_fk = sa.Column('ILT_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item = orm.relationship(StockInventory, backref=orm.backref('_taxes'))

    tax_fk = sa.Column('ILT_TAX_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    tax_cfk = sa.Column('ILT_TAX_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    tax = orm.relationship(Tax)


class DeliItem(Base):
    """
    Represents a deli item.
    """
    __tablename__ = 'DeliItems'
    __table_args__ = (
        sa.ForeignKeyConstraint(['DEL_INV_FK', 'DEL_INV_CFK'], ['StockInventory.INV_PK', 'StockInventory.INV_CPK']),
    )

    pk = sa.Column('DEL_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    cpk = sa.Column('DEL_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item_fk = sa.Column('DEL_INV_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    item_cfk = sa.Column('DEL_INV_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    item = orm.relationship(StockInventory, backref=orm.backref('deli', uselist=False))

    serving_size = sa.Column('DEL_ServingSize', sa.String(length=100), nullable=True)
    servings_per_container = sa.Column('DEL_ServingsPerContainer', sa.String(length=100), nullable=True)

    ingredients = sa.Column('DEL_IngredientList', sa.Text(), nullable=True)

    # nutrifacts = association_proxy('_nutrifacts', 'nutrifact')


class NutrifactItem(Base):
    """
    Represents an available nutrifact data point.
    """
    __tablename__ = 'NutrifactItems'

    pk = sa.Column('NUT_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('NUT_Description', sa.String(length=100), nullable=True)
    order = sa.Column('NUT_Order', sa.Integer(), nullable=True)


class DeliNutrifactItem(Base):
    """
    Represents a nutrifact data point for a deli item.
    """
    __tablename__ = 'NutrifactItemsLinkDeliItems'
    __table_args__ = (
        sa.ForeignKeyConstraint(['NLD_NUT_FK'], ['NutrifactItems.NUT_PK']),
        sa.ForeignKeyConstraint(['NLD_DEL_FK', 'NLD_DEL_CFK'], ['DeliItems.DEL_PK', 'DeliItems.DEL_CPK']),
    )

    nutrifact_fk = sa.Column('NLD_NUT_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    deli_fk = sa.Column('NLD_DEL_FK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    deli_cfk = sa.Column('NLD_DEL_CFK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    deli = orm.relationship(DeliItem, backref=orm.backref('nutrifacts'))
    nutrifact = orm.relationship(NutrifactItem)

    description = association_proxy('nutrifact', 'description')
    value = sa.Column('NLD_Value', sa.Numeric(precision=12, scale=3), nullable=True)
    percent = sa.Column('NLD_Percent', sa.Numeric(precision=12, scale=3), nullable=True)


@six.python_2_unicode_compatible
class Prefix(Base):
    """
    Represents a string "prefix".
    """
    __tablename__ = 'PrefixDescriptions'

    pk = sa.Column('PRE_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the prefix record.
    """)

    description = sa.Column('PRE_Description', sa.String(length=30), nullable=False, doc="""
    String value of the prefix.
    """)

    def __str__(self):
        return self.description or ''


@six.python_2_unicode_compatible
class Suffix(Base):
    """
    Represents a string "suffix".
    """
    __tablename__ = 'SuffixDescriptions'

    pk = sa.Column('SUF_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the suffix record.
    """)

    description = sa.Column('SUF_Description', sa.String(length=30), nullable=False, doc="""
    String value of the suffix.
    """)

    def __str__(self):
        return self.description or ''


class AddressDescription(Base):
    """
    Represents an "address description" record.
    """
    __tablename__ = 'ADescription'

    pk = sa.Column('ADE_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('ADE_Description', sa.String(length=30), nullable=False)


class Address(Base):
    """
    Represents a mailing address record.
    """
    __tablename__ = 'Addresses'
    __table_args__ = (
        sa.ForeignKeyConstraint(['ADD_ADE_FK'], ['ADescription.ADE_PK']),
        sa.ForeignKeyConstraint(['ADD_CUS_FK', 'ADD_CUS_CFK'], ['Customers.CUS_PK', 'Customers.CUS_CPK']),
    )

    pk = sa.Column('ADD_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    street1 = sa.Column('ADD_StreetAddressLine1', sa.String(length=30), nullable=True)

    street2 = sa.Column('ADD_StreetAddressLine2', sa.String(length=30), nullable=True)

    city = sa.Column('ADD_City', sa.String(length=20), nullable=True)

    state = sa.Column('ADD_StateProvince', sa.String(length=2), nullable=True)

    postal_code = sa.Column('ADD_PostalCode', sa.String(length=15), nullable=True)

    country = sa.Column('ADD_Country', sa.String(length=4), nullable=True)

    description_fk = sa.Column('ADD_ADE_FK', sa.Integer(), nullable=False)
    _description = orm.relationship(AddressDescription, uselist=False)
    description = association_proxy('_description', 'description')

    customer_fk = sa.Column('ADD_CUS_FK', sa.Integer(), nullable=True)
    customer_cfk = sa.Column('ADD_CUS_CFK', sa.Integer(), nullable=True)
    customer = orm.relationship('Customer',
        # primaryjoin='and_(Customer.pk == Address.customer_fk, Customer.cpk == Address.customer_cfk)',
        backref=orm.backref('addresses'))


class PhoneDescription(Base):
    """
    Represents an "phone description" record.
    """
    __tablename__ = 'PDescription'

    pk = sa.Column('PDE_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('PDE_Description', sa.String(length=30), nullable=False)


@six.python_2_unicode_compatible
class Phone(Base):
    """
    Represents a phone number record.
    """
    __tablename__ = 'PhoneNumbers'

    pk = sa.Column('PHO_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the phone record.
    """)

    number = sa.Column('PHO_PhoneNumber', sa.String(length=14), nullable=True, doc="""
    Phone number string.
    """)

    customer_fk = sa.Column('PHO_CUS_FK', sa.Integer(), sa.ForeignKey('Customers.CUS_PK'), nullable=True)

    customer_cfk = sa.Column('PHO_CUS_CFK', sa.Integer(), sa.ForeignKey('Customers.CUS_CPK'), nullable=True)

    customer = orm.relationship(
        'Customer',
        primaryjoin='and_(Customer.pk == Phone.customer_fk, Customer.cpk == Phone.customer_cfk)',
        backref=orm.backref('phones'))

    description_fk = sa.Column('PHO_PDE_FK', sa.Integer(), sa.ForeignKey('PDescription.PDE_PK'), nullable=True)

    _description = orm.relationship(
        PhoneDescription,
        uselist=False)

    description = association_proxy(
        '_description', 'description',
        creator=lambda d: PhoneDescription(description=d),
        getset_factory=getset_factory)

    modified = sa.Column('PHO_Timestamp', sa.DateTime(), nullable=True, doc="""
    Timestamp (without timezone, assumed to be "local") when the record was
    last modified.
    """)

    def __str__(self):
        return (self.number or '').strip()


class EmailDescription(Base):
    """
    Represents an "email description" record.
    """
    __tablename__ = 'EDescription'

    pk = sa.Column('EDE_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    description = sa.Column('EDE_Description', sa.String(length=30), nullable=False)


@six.python_2_unicode_compatible
class Email(Base):
    """
    Represents a email address record.
    """
    __tablename__ = 'EmailAddresses'

    pk = sa.Column('EMA_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the email record.
    """)

    address = sa.Column('EMA_EmailAddress', sa.String(length=50), nullable=True, doc="""
    Email address string.
    """)

    customer_fk = sa.Column('EMA_CUS_FK', sa.Integer(), sa.ForeignKey('Customers.CUS_PK'), nullable=True)

    customer_cfk = sa.Column('EMA_CUS_CFK', sa.Integer(), sa.ForeignKey('Customers.CUS_CPK'), nullable=True)

    customer = orm.relationship(
        'Customer',
        primaryjoin='and_(Customer.pk == Email.customer_fk, Customer.cpk == Email.customer_cfk)',
        backref=orm.backref('emails'))

    description_fk = sa.Column('EMA_EDE_FK', sa.Integer(), sa.ForeignKey('EDescription.EDE_PK'), nullable=True)

    _description = orm.relationship(
        EmailDescription,
        uselist=False)

    description = association_proxy(
        '_description', 'description',
        creator=lambda d: EmailDescription(description=d),
        getset_factory=getset_factory)

    receipt_optin_date = sa.Column('EMA_eReceiptOptInDate', sa.DateTime(), nullable=True)

    modified = sa.Column('EMA_TimeStamp', sa.DateTime(), nullable=True, doc="""
    Timestamp (without timezone, assumed to be "local") when the record was
    last modified.
    """)

    def __str__(self):
        return (self.address or '').strip()


class Employee(Base):
    """
    Represents an employee record.
    """
    __tablename__ = 'Employees'

    pk = sa.Column('EMP_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    first_name = sa.Column('EMP_FirstName', sa.String(length=30), nullable=False)

    middle_name = sa.Column('EMP_MiddleName', sa.String(length=30), nullable=True)

    last_name = sa.Column('EMP_LastName', sa.String(length=30), nullable=False)


@six.python_2_unicode_compatible
class Customer(Base):
    """
    Represents a customer record.
    """
    __tablename__ = 'Customers'

    pk = sa.Column('CUS_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False, doc="""
    Primary key for the customer record.
    """)

    cpk = sa.Column('CUS_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    account_number = sa.Column('CUS_AccountNum', sa.String(length=16), nullable=False, doc="""
    Presumably unique account number for the customer.
    """)

    prefix_fk = sa.Column('CUS_PRE_FK', sa.Integer(), sa.ForeignKey('PrefixDescriptions.PRE_PK'), nullable=True, doc="""
    Foreign key of the customer's name prefix, if any.
    """)

    _prefix = orm.relationship(Prefix, doc="""
    Reference to the :class:`Prefix` instance associated with the customer, if any.
    """)

    prefix = association_proxy(
        '_prefix', 'description',
        creator=lambda p: Prefix(description=p),
        getset_factory=getset_factory)

    first_name = sa.Column('CUS_FirstName', sa.String(length=25), nullable=True, doc="""
    First name for the customer.
    """)

    middle_name = sa.Column('CUS_MiddleName', sa.String(length=25), nullable=True, doc="""
    Middle name for the customer.
    """)

    last_name = sa.Column('CUS_LastName', sa.String(length=25), nullable=True, doc="""
    Last name for the customer.
    """)

    suffix_fk = sa.Column('CUS_SUF_FK', sa.Integer(), sa.ForeignKey('SuffixDescriptions.SUF_PK'), nullable=True, doc="""
    Foreign key of the customer's name suffix, if any.
    """)

    _suffix = orm.relationship(Suffix, doc="""
    Reference to the :class:`Suffix` instance associated with the customer, if any.
    """)

    suffix = association_proxy(
        '_suffix', 'description',
        creator=lambda p: Suffix(description=p),
        getset_factory=getset_factory)

    title = sa.Column('CUS_Title', sa.String(length=25), nullable=True, doc="""
    "Title" string for the customer.
    """)

    company = sa.Column('CUS_Company', sa.String(length=25), nullable=True, doc="""
    "Company" string for the customer.
    """)

    default_price_level = sa.Column('CUS_DefaultPriceLevel', sa.Integer(), nullable=True, doc="""
    Default price level code for the customer.
    """)

    suppress_receipt = sa.Column('CUS_SuppressReceipt', sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the customer wishes to opt out of printed receipts.
    """)

    memo = sa.Column('CUS_Memo', sa.String(length=254), nullable=True)

    hide_from_lookup = sa.Column('CUS_HideInLookup', sa.Boolean(), nullable=True)

    automatic_discount_fk = sa.Column('CUS_DIS_FK', sa.Integer(), sa.ForeignKey('DiscountProfiles.DIS_PK'), nullable=True)

    automatic_discount = orm.relationship(DiscountProfile)

    inactive = sa.Column('CUS_InactiveFlag', sa.Boolean(), nullable=False, doc="""
    Inactive flag for the customer.
    """)

    modified = sa.Column('CUS_TimeStamp', sa.DateTime(), nullable=True, doc="""
    Timestamp (without timezone, assumed to be "local") when the record was
    last modified.
    """)

    store_coupons = association_proxy(
        '_store_coupons', 'coupon',
        creator='lambda c: CustomerStoreCoupon(coupon=c)',
        getset_factory=getset_factory)

    def __str__(self):
        return "{} {} {}".format(self.account_number.strip(),
                                 (self.first_name or '').strip(),
                                 (self.last_name or '').strip())


class StoreCouponProfile(Base):
    """
    Represents a store coupon profile record.
    """
    __tablename__ = 'StoreCouponProfiles'

    pk = sa.Column('SCP_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    name = sa.Column('SCP_Name', sa.String(length=30), nullable=False)

    _customers = orm.relationship(
        'CustomerStoreCoupon',
        backref=orm.backref('coupon'))

    customers = association_proxy(
        '_customers', 'customer',
        creator='lambda c: CustomerStoreCoupon(customer=c)',
        getset_factory=getset_factory)


class CustomerStoreCoupon(Base):
    """
    Represents the link between a customer and store coupon profile.
    """
    __tablename__ = 'CustomerLinkStoreCouponProfiles'

    customer_fk = sa.Column('CLS_CUS_FK', sa.Integer(), sa.ForeignKey('Customers.CUS_PK'), primary_key=True, autoincrement=False, nullable=False)

    customer_cfk = sa.Column('CLS_CUS_CFK', sa.Integer(), sa.ForeignKey('Customers.CUS_CPK'), primary_key=True, autoincrement=False, nullable=False)

    customer = orm.relationship(
        Customer,
        primaryjoin='and_(Customer.pk == CustomerStoreCoupon.customer_fk, Customer.cpk == CustomerStoreCoupon.customer_cfk)',
        backref=orm.backref('_store_coupons'))

    coupon_fk = sa.Column('CLS_SCP_FK', sa.Integer(), sa.ForeignKey('StoreCouponProfiles.SCP_PK'), primary_key=True, autoincrement=False, nullable=False)


@six.python_2_unicode_compatible
class WorksheetType(Base):
    """
    Represents a worksheet type.
    """
    __tablename__ = 'WorksheetTypes'

    pk = sa.Column('WST_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    name = sa.Column('WST_Name', sa.String(length=30), nullable=False)

    db_table = sa.Column('WST_DBTable', sa.String(length=40), nullable=True)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Worksheet(Base):
    """
    Represents a worksheet.
    """
    __tablename__ = 'Worksheets'

    pk = sa.Column('WRK_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    type_fk = sa.Column('WRK_WST_FK', sa.Integer(), sa.ForeignKey('WorksheetTypes.WST_PK'), nullable=False)

    type = orm.relationship(WorksheetType)

    name = sa.Column('WRK_Name', sa.String(length=40), nullable=False)

    start_time = sa.Column('WRK_StartDate', sa.DateTime(), nullable=True)

    end_time = sa.Column('WRK_EndDate', sa.DateTime(), nullable=True)

    status = sa.Column('WRK_Status', sa.String(length=30), nullable=True)

    timestamp = sa.Column('WRK_Timestamp', sa.DateTime(), nullable=True)

    timestamp_committed = sa.Column('WRK_TimestampCommitted', sa.DateTime(), nullable=True)

    committed = sa.Column('WRK_Committed', sa.Boolean(), nullable=True)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class PromotionalSaleWorksheet(Base):
    """
    Represents a promotional sale worksheet.
    """
    __tablename__ = 'PromotionalSaleWorkSheet'

    pk = sa.Column('PSW_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('PSW_WRK_FK', sa.Integer(), sa.ForeignKey('Worksheets.WRK_PK'), nullable=False)

    worksheet = orm.relationship(
        Worksheet,
        backref=orm.backref('promo_sale', uselist=False))

    start_time = sa.Column('PSW_StartDate', sa.DateTime(), nullable=True)

    end_time = sa.Column('PSW_EndDate', sa.DateTime(), nullable=True)

    permanent = sa.Column('PSW_Permanent', sa.Boolean(), nullable=True)

    priority_level = sa.Column('PSW_PriorityLevel', sa.Integer(), nullable=True)

    def __str__(self):
        return (self.worksheet.name or '') if self.worksheet else ''


class PromotionalSaleWorksheetItem(Base):
    """
    Represents a "line item" within a promotional sale worksheet.
    """
    __tablename__ = 'PromotionalSaleWorksheetData'
    __table_args__ = (
        sa.ForeignKeyConstraint(['PSD_PSW_FK'], ['PromotionalSaleWorkSheet.PSW_PK']),
        sa.ForeignKeyConstraint(['PSD_INV_FK'], ['v_InventoryMaster.inv_pk']),
        sa.ForeignKeyConstraint(['PSD_DIS_FK_pl1'], ['DiscountProfiles.DIS_PK']),
    )

    pk = sa.Column('PSD_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('PSD_PSW_FK', sa.Integer(), nullable=True)
    worksheet = orm.relationship(PromotionalSaleWorksheet, backref=orm.backref('items'))

    item_fk = sa.Column('PSD_INV_FK', sa.Integer(), nullable=True)
    item = orm.relationship(InventoryItem, backref=orm.backref('_promo_sale_items'))

    promo_price1 = sa.Column('PSD_PromotionalPrice_pl1', sa.Numeric(precision=16, scale=4), nullable=True)

    discount_pl1_fk = sa.Column('PSD_DIS_FK_pl1', sa.Integer(), nullable=True)
    discount_pl1 = orm.relationship(DiscountProfile)


@six.python_2_unicode_compatible
class InventoryLabelWorksheet(Base):
    """
    Represents an inventory label worksheet.
    """
    __tablename__ = 'InventoryLabelWorksheets'

    pk = sa.Column('ILW_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('ILW_WRK_FK', sa.Integer(), sa.ForeignKey('Worksheets.WRK_PK'), nullable=False)

    worksheet = orm.relationship(
        Worksheet,
        backref=orm.backref('inv_labels', uselist=False))

    def __str__(self):
        return (self.worksheet.name or '') if self.worksheet else ''


class InventoryLabelWorksheetItem(Base):
    """
    Represents a "line item" within an inventory label worksheet.
    """
    __tablename__ = 'InventoryLabelWorksheetData'

    pk = sa.Column('IWD_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('IWD_ILW_FK', sa.Integer(), sa.ForeignKey('InventoryLabelWorksheets.ILW_PK'), nullable=False)

    worksheet = orm.relationship(
        InventoryLabelWorksheet,
        backref=orm.backref('items'))

    item_fk = sa.Column('IWD_INV_FK', sa.Integer(), sa.ForeignKey('StockInventory.INV_PK'), nullable=False)

    item = orm.relationship(StockInventory)

    label_quantity = sa.Column('IWD_LabelQty', sa.String(length=3), nullable=True)


@six.python_2_unicode_compatible
class AddItemsWorksheet(Base):
    """
    Represents an "add items" worksheet.
    """
    __tablename__ = 'AddItemsWorksheet'

    pk = sa.Column('AIW_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('AIW_WRK_FK', sa.Integer(), sa.ForeignKey('Worksheets.WRK_PK'), nullable=False)

    worksheet = orm.relationship(
        Worksheet,
        backref=orm.backref('add_items', uselist=False))

    def __str__(self):
        return (self.worksheet.name or '') if self.worksheet else ''


class AddItemsWorksheetItem(Base):
    """
    Represents a "line item" within an "add items" worksheet.
    """
    __tablename__ = 'AddItemsWorksheetData'

    pk = sa.Column('AID_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    worksheet_fk = sa.Column('AID_AIW_FK', sa.Integer(), sa.ForeignKey('AddItemsWorksheet.AIW_PK'), nullable=False)

    worksheet = orm.relationship(
        AddItemsWorksheet,
        backref=orm.backref('items'))

    item_id = sa.Column('AID_ItemID', sa.String(length=14), nullable=True)

    item = orm.relationship(
        StockInventory,
        primaryjoin=StockInventory.scancode == item_id,
        foreign_keys=[item_id])

    alternate = orm.relationship(
        ItemAlternate,
        primaryjoin=ItemAlternate.scancode == item_id,
        foreign_keys=[item_id])

    receipt_alias = sa.Column('AID_ReceiptAlias', sa.String(length=32), nullable=True)

    brand = sa.Column('AID_Brand', sa.String(length=30), nullable=True)

    name = sa.Column('AID_Name', sa.String(length=30), nullable=True)

    size = sa.Column('AID_Size', sa.String(length=20), nullable=True)


class Transaction(Base):
    """
    Represents a (presumably POS) sale transaction.
    """
    __tablename__ = 'Transactions'

    pk = sa.Column('TRN_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    
    cpk = sa.Column('TRN_CPK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)
    
    start_time = sa.Column('TRN_StartTime', sa.DateTime(), nullable=False)

    end_time = sa.Column('TRN_EndTime', sa.DateTime(), nullable=False)

    upload_time = sa.Column('TRN_UploadTime', sa.DateTime(), nullable=False)

    receipt_number = sa.Column('TRN_ReceiptNumber', sa.Integer(), nullable=False)

    employee_fk = sa.Column('TRN_EMP_FK', sa.Integer(), sa.ForeignKey('Employees.EMP_PK'), nullable=False)

    employee = orm.relationship(
        Employee,
        backref=orm.backref('transactions'))

    employee_first_name = association_proxy(
        'employee', 'first_name',
        creator='lambda n: Employee(first_name=n)',
        getset_factory=getset_factory)

    employee_last_name = association_proxy(
        'employee', 'last_name',
        creator='lambda n: Employee(last_name=n)',
        getset_factory=getset_factory)

    customer_fk = sa.Column('TRN_CUS_FK', sa.Integer(), sa.ForeignKey('Customers.CUS_PK'), nullable=True)

    customer_cfk = sa.Column('TRN_CUS_CFK', sa.Integer(), sa.ForeignKey('Customers.CUS_CPK'), nullable=False)

    customer = orm.relationship(
        Customer,
        primaryjoin='and_(Customer.pk == Transaction.customer_fk, Customer.cpk == Transaction.customer_cfk)',
        backref=orm.backref('transactions'))

    account_number = association_proxy(
        'customer', 'account_number',
        creator='lambda n: Customer(account_number=n)',
        getset_factory=getset_factory)

    status = sa.Column('TRN_Status', sa.Integer(), nullable=True)

    all_void = sa.Column('TRN_AllVoid', sa.Boolean(), nullable=False)

    net_sales = sa.Column('TRN_NetSales', sa.Numeric(precision=16, scale=4), nullable=True)

    total = sa.Column('TRN_Total', sa.Numeric(precision=16, scale=4), nullable=True)


class TransactionLineItem(Base):
    """
    Represents a line item from a sale transaction.
    """
    __tablename__ = 'TransactionLineItems'

    transaction_fk = sa.Column('TLI_TRN_FK', sa.Integer(), sa.ForeignKey('Transactions.TRN_PK'), primary_key=True, autoincrement=False, nullable=False)

    transaction_cfk = sa.Column('TLI_TRN_CFK', sa.Integer(), sa.ForeignKey('Transactions.TRN_CPK'), primary_key=True, autoincrement=False, nullable=False)

    pk = sa.Column('TLI_PK', sa.Integer(), primary_key=True, autoincrement=False, nullable=False)

    transaction = orm.relationship(
        Transaction,
        primaryjoin='and_(Transaction.pk == TransactionLineItem.transaction_fk, Transaction.cpk == TransactionLineItem.transaction_cfk)',
        backref=orm.backref('line_items'))

    transaction_sequence_number = sa.Column('TLI_TransactionSequenceNumber', sa.Integer(), nullable=False)

    type_fk = sa.Column('TLI_LIT_FK', sa.Integer(), nullable=True)
    
    start_time = sa.Column('TLI_StartTime', sa.DateTime(), nullable=False)

    end_time = sa.Column('TLI_EndTime', sa.DateTime(), nullable=False)

    scancode = sa.Column('TLI_ScanCode', sa.String(length=14), nullable=True)

    receipt_alias = sa.Column('TLI_ReceiptAlias', sa.String(length=33), nullable=True)

    quantity = sa.Column('TLI_Quantity', sa.Numeric(precision=12, scale=3), nullable=True)

    amount = sa.Column('TLI_Amount', sa.Numeric(precision=16, scale=4), nullable=True)

    extended_price = sa.Column('TLI_ExtendedPrice', sa.Numeric(precision=16, scale=4), nullable=True)

    tax_difference = sa.Column('TLI_TaxDifference', sa.Numeric(precision=16, scale=4), nullable=False)


class InventoryTransactionItem(Base):
    """
    Represents the "inventory" aspects for a sale transaction line item.
    """
    __tablename__ = 'InventoryTransactionItems'

    transaction_fk = sa.Column('ITI_TRN_FK', sa.Integer(), sa.ForeignKey('Transactions.TRN_PK'), primary_key=True, autoincrement=False, nullable=False)

    transaction_cfk = sa.Column('ITI_TRN_CFK', sa.Integer(), sa.ForeignKey('Transactions.TRN_CPK'), primary_key=True, autoincrement=False, nullable=False)

    line_item_fk = sa.Column('ITI_TLI_FK', sa.Integer(), sa.ForeignKey('TransactionLineItems.TLI_PK'), primary_key=True, autoincrement=False, nullable=False)

    transaction = orm.relationship(
        Transaction,
        primaryjoin='and_(Transaction.pk == InventoryTransactionItem.transaction_fk, Transaction.cpk == InventoryTransactionItem.transaction_cfk)',
        backref=orm.backref('inventory_line_items'))

    line_item = orm.relationship(
        TransactionLineItem,
        backref=orm.backref(
            'inventory',
            uselist=False,
            lazy='joined'))

    item_fk = sa.Column('ITI_INV_FK', sa.Integer(), sa.ForeignKey('v_InventoryMaster.inv_pk'), nullable=True)

    item = orm.relationship(
        InventoryItem,
        backref=orm.backref('_inventory_transaction_items'))

    scancode = sa.Column('ITI_ScanCode', sa.String(length=14), nullable=True)

    quantity_sold = sa.Column('ITI_QuantitySold', sa.Numeric(precision=12, scale=3), nullable=False)

    selling_price = sa.Column('ITI_SellingPrice', sa.Numeric(precision=16, scale=4), nullable=False)


class DiscountTransactionItem(Base):
    """
    Represents a transaction "discount" line item.
    """
    __tablename__ = 'DiscountTransactionItems'

    transaction_fk = sa.Column('DLI_TRN_FK', sa.Integer(), sa.ForeignKey('Transactions.TRN_PK'), primary_key=True, autoincrement=False, nullable=False)

    transaction_cfk = sa.Column('DLI_TRN_CFK', sa.Integer(), sa.ForeignKey('Transactions.TRN_CPK'), primary_key=True, autoincrement=False, nullable=False)

    line_item_fk = sa.Column('DLI_TLI_FK', sa.Integer(), sa.ForeignKey('TransactionLineItems.TLI_PK'), primary_key=True, autoincrement=False, nullable=False)

    transaction = orm.relationship(
        Transaction,
        primaryjoin='and_(Transaction.pk == DiscountTransactionItem.transaction_fk, Transaction.cpk == DiscountTransactionItem.transaction_cfk)',
        backref=orm.backref('discount_line_items'))

    line_item = orm.relationship(
        TransactionLineItem,
        backref=orm.backref(
            'discounts',
            lazy='joined'))

    discount_fk = sa.Column('DLI_DIS_FK', sa.Integer(), sa.ForeignKey('DiscountProfiles.DIS_PK'), nullable=False)

    discount = orm.relationship(
        DiscountProfile,
        backref=orm.backref('_discount_transaction_items'))

    amount = sa.Column('DLI_Amount', sa.Numeric(precision=16, scale=4), nullable=False)

    inventory_line_item = sa.Column('DLI_InventoryLineItem', sa.Integer(), nullable=True)
