# -*- coding: utf-8 -*-
################################################################################
#
#  pyCatapult -- Python Interface to ECRS Catapult
#  Copyright © 2015-2016 Lance Edgar
#
#  This file is part of pyCatapult.
#
#  pyCatapult is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  pyCatapult is distributed in the hope that it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  pyCatapult.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Catapult Web API
"""

from __future__ import unicode_literals, absolute_import

import datetime
import decimal
from xml.etree import ElementTree

import requests

from catapult.webapi import model


class SQLError(Exception):
    """
    'SQLerror' is the name of the root node of the XML document returned by the
    Catapult Web API whenever something goes awry, hence this class.
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return b"Catapult REST API returned an error: {}".format(self.message)


class CatapultWebAPI(object):
    """
    Generic web API object.
    """

    def __init__(self, base_url, apikey, verify=True):
        self.apikey = apikey
        self.verify = verify

        # Make sure URL ends with /api (good idea?)
        self.base_url = base_url.rstrip('/')
        if not self.base_url.endswith('/api'):
            self.base_url += '/api'

    def post(self, method, data):
        response = requests.post('{}/{}'.format(self.base_url, method), params=data, verify=self.verify)
        response.raise_for_status()
        return response

    def put(self, method, data):
        response = requests.put('{}/{}'.format(self.base_url, method), params=data, verify=self.verify)
        response.raise_for_status()
        return response

    def delete(self, method, data):
        response = requests.delete('{}/{}'.format(self.base_url, method), params=data, verify=self.verify)
        response.raise_for_status()
        return response

    def get_customer(self, customerId=None, **kwargs):
        """
        Fetch an existing customer record from Catapult.
        """
        data = {'apikey': self.apikey}
        if customerId is not None:
            data['customerId'] = unicode(customerId)
        data.update(kwargs)
        response = self.post('Customer', data)
        return self.parse_customer(response)

    def force_empty_values(self, data, fields):
        for field in fields:
            if field in data and not data[field]:
                data[field] = ' '

    def put_customer(self, **kwargs):
        """
        Create a new customer record in Catapult.
        """
        data = {'apikey': self.apikey}
        data.update(kwargs)

        # TODO: at least for now (?), must send space instead of empty string, to
        # force value change for some fields, in case was previously non-empty
        self.force_empty_values(data, [
            'middleName',
            'automaticDiscount',
            'billToEmailAddress',
            'billToPhoneNumber',
        ])

        response = self.put('Customer', data)
        return self.parse_customer(response)

    def parse_customer(self, response):
        """
        Parse a response containing XML, returning a more Pythonic ``Customer``
        object with various attributes set.
        """
        root = ElementTree.fromstring(response.content)
        if root.tag == 'SQLerror':
            raise SQLError(root.attrib['message'])
        assert len(root) == 1, "Expected exactly 1 match, got {} instead".format(len(root))
        row = root[0]

        customer = model.Customer()
        customer.account_number = row.attrib['customerId']
        customer.first_name = row.attrib['firstName']
        customer.middle_name = row.attrib['middleName']
        customer.last_name = row.attrib['lastName']
        customer.prefix = row.attrib['prefix']
        customer.suffix = row.attrib['suffix']
        customer.memo = row.attrib['memo']
        customer.price_level_name = row.attrib['priceLevel']
        customer.suppress_receipt = row.attrib['suppressReceipt'] == 'true'
        customer.hide_from_lookup = row.attrib['hideInLookup'] == 'true'
        customer.power_field_1 = row.attrib['powerField1']
        customer.power_field_2 = row.attrib['powerField2']
        customer.power_field_3 = row.attrib['powerField3']
        customer.power_field_4 = row.attrib['powerField4']
        customer.power_field_5 = row.attrib['powerField5']
        customer.power_field_6 = row.attrib['powerField6']
        customer.power_field_7 = row.attrib['powerField7']
        customer.power_field_8 = row.attrib['powerField8']
        customer.birth_date = row.attrib['birthDate']
        customer.membership_account_name = row.attrib['membershipAccount']
        customer.check_acceptance_profile = row.attrib['checkAcceptanceProfile']
        customer.automatic_discount_description = row.attrib['automaticDiscount']
        customer.loyalty_enabled = row.attrib['loyaltyEnabled'] == 'true'
        customer.loyalty_accrual_multiplier = row.attrib['loyaltyAccrualMultiplier']
        customer.check_file_memo = row.attrib['checkFileMemo']

        customer.automatic_discount = None
        discount_name = row.attrib['automaticDiscount']
        if discount_name:
            discount = model.DiscountProfile()
            discount.description = discount_name
            customer.automatic_discount = discount

        address = model.Address()
        address.street1 = row.attrib['billToAddressLine1']
        address.street2 = row.attrib['billToAddressLine2']
        address.city = row.attrib['billToAddressCity']
        address.state = row.attrib['billToAddressStateProv']
        address.postal_code = row.attrib['billToAddressPostalCode']
        address.description = row.attrib['billToAddressDescription']
        customer.addresses = [address]

        customer.emails = []
        email_address = row.attrib['billToEmailAddress']
        if email_address:
            email = model.Email()
            email.address = email_address
            email.description = row.attrib['billToEmailDescription']
            customer.emails.append(email)

        customer.phones = []
        phone_number = row.attrib['billToPhoneNumber']
        if phone_number:
            phone = model.Phone()
            phone.number = phone_number
            phone.description = row.attrib['billToPhoneDescription']
            customer.phones.append(phone)

        return customer

    def get_membership_profile(self, name=None, **kwargs):
        """
        Fetch an existing membership profile record from Catapult.
        """
        data = {'apikey': self.apikey}
        if name is not None:
            data['membershipProfile'] = unicode(name)
        data.update(kwargs)
        response = self.post('MembershipProfile', data)
        return self.parse_membership_profile(response)

    def parse_membership_profile(self, response):
        """
        Parse a response containing XML, returning a more Pythonic
        ``MembershipProfile`` object with various attributes set.
        """
        root = ElementTree.fromstring(response.content)
        if root.tag == 'SQLerror':
            raise SQLError(root.attrib['message'])
        assert len(root) == 1, "Expected exactly 1 match, got {} instead".format(len(root))
        row = root[0]

        profile = model.MembershipProfile()
        profile.name = row.attrib['membershipProfile']
        profile.is_private = row.attrib['isPrivate']
        profile.is_default = row.attrib['isDefault']
        profile.automatic_discount_description = row.attrib['automaticDiscount']
        profile.price_level_name = row.attrib['priceLevel']
        profile.equity_enabled = row.attrib['equityEnabled']
        profile.max_equity = row.attrib['maxEquity']
        profile.equity_mode = row.attrib['equityMode']
        profile.equity_payment_amount = row.attrib['equityPaymentAmount']
        profile.equity_payment_interval = row.attrib['equityPaymentInterval']
        profile.surcharge_percent = row.attrib['surchargePercent']
        profile.membership_fee_enabled = row.attrib['membershipFeeEnabled']
        profile.membership_fee_amount = row.attrib['membershipFeeAmount']
        profile.membership_fee_interval = row.attrib['membershipFeeInterval']
        profile.join_fee_enabled = row.attrib['joinFeeEnabled']
        profile.join_fee_amount = row.attrib['joinFeeAmount']
        return profile

    def put_customer_store_coupon(self, **kwargs):
        """
        Attach a store coupon to a customer.
        """
        data = {'apikey': self.apikey}
        data.update(kwargs)
        response = self.put('CustomerStoreCoupon', data)
        return self.parse_customer_store_coupons(response)

    def delete_customer_store_coupon(self, **kwargs):
        """
        Remove a store coupon from a customer.
        """
        data = {'apikey': self.apikey}
        data.update(kwargs)
        response = self.delete('CustomerStoreCoupon', data)
        return response

    def parse_customer_store_coupons(self, response):
        """
        Parse a response containing XML, returning a more Pythonic sequence of
        ``CustomerStoreCoupon`` objects with various attributes set.
        """
        root = ElementTree.fromstring(response.content)
        if root.tag == 'SQLerror':
            raise SQLError(root.attrib['message'])

        coupons = []
        for row in root:
            coupon = model.CustomerStoreCoupon()
            coupon.customer_account_number = row.attrib['customerId']
            coupon.name = row.attrib['couponName']
            coupon.code = row.attrib['couponCode']
            coupon.receipt_alias = row.attrib['couponReceiptAlias']
            coupon.expires = row.attrib['couponExpires']
            coupons.append(coupon)
        return coupons

    def get_transactions(self, startTime=None, endTime=None, **kwargs):
        """
        Fetch all POS transactions within the given time range.  Note that the
        start and end times are assumed to be in the same time zone as the
        Catapult server / data.
        """
        data = {'apikey': self.apikey}
        if startTime is not None:
            data['startTime'] = startTime.isoformat()
        if endTime is not None:
            data['endTime'] = endTime.isoformat()
        data.update(kwargs)
        response = self.post('TransactionJournal', data)
        return self.parse_transactions(response)

    def parse_transactions(self, response):
        root = ElementTree.fromstring(response.content)
        if root.tag == 'SQLerror':
            raise SQLError(root.attrib['message'])

        # first group all rows according to transaction
        grouped = {}
        for row in root:
            pk = int(row.attrib['TRN_PK'])
            grouped.setdefault(pk, []).append(row)

        # next make sure all rows are in sequence, for each transaction
        for pk in list(grouped):
            rows = grouped[pk]
            grouped[pk] = sorted(rows, key=lambda row: int(row.attrib.get('TLI_PK', 0)))

        # okay now we can mock up some transaction objects...
        transactions = []
        for pk, rows in grouped.iteritems():

            row = rows[0]
            transaction = model.Transaction()
            transaction._row_element = row
            transaction.pk = pk
            transaction.all_void = row.attrib['TRN_AllVoid'] == '1'
            transaction.start_time = datetime.datetime.strptime(row.attrib['TRN_StartTime'], '%Y-%m-%d %H:%M:%S.%f')
            transaction.end_time = datetime.datetime.strptime(row.attrib['TRN_EndTime'], '%Y-%m-%d %H:%M:%S.%f')
            transaction.upload_time = datetime.datetime.strptime(row.attrib['TRN_UploadTime'], '%Y-%m-%d %H:%M:%S.%f')
            transaction.terminal_number = int(row.attrib.get('TCF_Number', 0)) or None
            transaction.receipt_number = int(row.attrib['TRN_ReceiptNumber'])
            transaction.account_number = row.attrib.get('CUS_AccountNum', '').strip() or None
            transaction.employee_first_name = row.attrib['EMP_FirstName'].strip()
            transaction.employee_last_name = row.attrib['EMP_LastName'].strip()
            transaction.type = row.attrib['TRN_Type']
            transaction.status = row.attrib['TRN_Status']
            transaction.training_mode = transaction.status == '1'
            transaction.loan = row.attrib.get('NTI_Loan') == '1'
            transaction.safe_drop = row.attrib.get('NTI_SafeDrop') == '1'
            transaction.no_sale = len(rows) == 1 and int(row.attrib.get('TLI_LIT_FK', 0)) == 6 # No Sale
            transaction.reconciliation = len(rows) == 1 and int(row.attrib.get('TLI_LIT_FK', 0)) == 21 # Reconciliation Action
            transaction.tax = 0
            transaction.line_items = []

            for row in rows:
                type_fk = int(row.attrib.get('TLI_LIT_FK', 0))
                receipt_alias = row.attrib.get('TLI_ReceiptAlias', '')
                amount = decimal.Decimal(row.attrib.get('TLI_Amount', '0'))

                if type_fk == 3: # Tax
                    transaction.tax += amount

                elif type_fk == 14: # Totals Entry
                    if receipt_alias == 'SUBTOTAL':
                        transaction.subtotal = amount
                    elif receipt_alias == 'TOTAL':
                        transaction.total = amount

                elif type_fk:
                    item = model.TransactionLineItem()
                    item._row_element = row
                    item.pk = int(row.attrib.get('TLI_PK', 0)) or None
                    item.type_fk = type_fk
                    item.void = row.attrib.get('ITI_Void', '0') == '1'
                    item.returned = type_fk == 15 # Returned Merchandise
                    item.start_time = row.attrib.get('TLI_StartTime', '')
                    item.end_time = row.attrib.get('TLI_EndTime', '')
                    item.amount = amount
                    item.quantity = decimal.Decimal(row.attrib.get('TLI_Quantity', '0'))
                    item.scancode = row.attrib.get('TLI_ScanCode', '')
                    item.receipt_alias = receipt_alias
                    item.department_number = int(row.attrib.get('DPT_Number', 0)) or None
                    item.discount_fk = int(row.attrib.get('DLI_DIS_FK', 0)) or None
                    if item.discount_fk:
                        item.discounted_line_number = int(row.attrib.get('DLI_InventoryLineItem', 0)) or None
                    item.tax_difference = decimal.Decimal(row.attrib.get('TLI_TaxDifference', '0'))
                    item.extended_price = decimal.Decimal(row.attrib.get('ExtendedPrice', '0'))
                    # item.extended_base_price = decimal.Decimal(row.attrib['ExtendedBasePrice'])
                    transaction.line_items.append(item)

            transactions.append(transaction)
        return transactions
